@@ -1,11 +1,14 @@
 """Incremental import of git into darcs."""
 
+
 import os
 import sys
+import xml.etree.ElementTree as ET
+from collections import OrderedDict
 from contextlib import contextmanager
 from datetime import datetime
 from pathlib import Path
-from shutil import rmtree
+from shutil import copy, rmtree
 from subprocess import DEVNULL, PIPE, CalledProcessError
 from subprocess import Popen as SPOpen
 from subprocess import run as srun
@@ -13,6 +16,8 @@ from threading import Thread
 
 import click
 from click import ClickException
+from colorama import Fore, Style, init
+from readchar import readkey
 from tqdm import tqdm
 
 _uuid = "_b531990e-3187-4b52-be1f-6e4d4d1e40c9"
@@ -23,6 +28,25 @@ _verbose = False
 _devnull = DEVNULL
 _disable = None
 _shutdown = False
+_darcs_date = "%Y%m%d%H%M%S"
+_pull_question = "Shall I pull this patch"
+_pull_help = """
+y: pull this patch
+n: don't pull it
+w: decide later
+
+a: pull all remaining patches
+i: don't pull remaining patches
+
+l: show full log message
+f: show full patch
+
+?: help
+h: help
+
+c: cancel without pulling
+q: cancel without pulling
+"""
 
 _boring = """
 # git
@@ -31,6 +55,8 @@ _boring = """
 (^|/)_darcs($|/)
 """
 
+_gitignore = "/_darcs"
+
 
 def handle_shutdown():
     """Wait for CTRL-D and set _shutdown, to flag a graceful shutdown request."""
@@ -42,11 +68,20 @@ def handle_shutdown():
     _shutdown = True
 
 
+def args_print(args):
+    """Print args of executed command."""
+    if _verbose:
+        args = [str(x) for x in args]
+        args = " ".join(args)
+        print(f"\n$ {args}")
+
+
 class Popen(SPOpen):
     """Inject defaults into Popen."""
 
     def __init__(self, *args, stderr=None, stdin=None, **kwargs):
         """Inject default into Popen."""
+        args_print(args[0])
         if not stderr:
             stderr = _devnull
         if not stdin and "input" not in kwargs:
@@ -56,6 +91,7 @@ class Popen(SPOpen):
 
 def run(*args, stdout=None, stderr=None, stdin=None, **kwargs):
     """Inject defaults into run."""
+    args_print(args[0])
     if not stdout:
         stdout = _devnull
     if not stderr:
@@ -65,6 +101,15 @@ def run(*args, stdout=None, stderr=None, stdin=None, **kwargs):
     return srun(*args, stdout=stdout, stderr=stderr, stdin=stdin, **kwargs)
 
 
+def hasnew():
+    """Revert recorded changes in darcs."""
+    try:
+        run(["darcs", "whatsnew"], check=True)
+    except CalledProcessError:
+        return False
+    return True
+
+
 def revert():
     """Revert recorded changes in darcs."""
     run(["darcs", "revert", "--no-interactive"])
@@ -93,14 +138,14 @@ def move(orig, new):
     if (porig.is_file() or porig.is_dir()) and not porig.is_symlink():
         dir = Path(new).parent
         dir.mkdir(parents=True, exist_ok=True)
-        add(dir)
+        darcs_add(dir)
         run(
             ["darcs", "move", "--case-ok", "--reserved-ok", orig, new],
             check=True,
         )
 
 
-def add(path):
+def darcs_add(path):
     """Add a path to the darcs-repo."""
     try:
         run(
@@ -134,7 +179,64 @@ def get_tags():
 
 def darcs_clone(source, destination):
     """Clone darcs-repo."""
-    run(["darcs", "clone", source, destination], check=True)
+    run(
+        ["darcs", "clone", source, destination],
+        check=True,
+    )
+
+
+def get_patches(source, args):
+    """Get patches from darcs."""
+    res = run(
+        ["darcs", "pull", "--dry-run", "--xml-output"] + args + [source],
+        stdout=PIPE,
+        check=True,
+    )
+    res = res.stdout.decode("UTF-8").strip()
+    if res.startswith("No remote patches to pull in!"):
+        return ET.fromstring("<root></root>")
+    else:
+        return ET.fromstring(res)
+
+
+def show_full_patch(source, patch):
+    """Show full patch with darcs."""
+    run(
+        ["darcs", "log", "-i", "--repodir", source, "-h", patch],
+        stdout=sys.stdout,
+        stderr=sys.stderr,
+        input=b"y",
+        check=True,
+    )
+
+
+def pull_patch(source, hash):
+    """Pull a darcs-patch."""
+    run(
+        [
+            "darcs",
+            "pull",
+            "--no-interactive",
+            "--no-set-default",
+            source,
+            "-h",
+            hash,
+        ],
+        check=True,
+    )
+
+
+def git_add(args=None):
+    """Add changes to git."""
+    if args is None:
+        run(["git", "add", "."], check=True)
+    else:
+        run(["git", "add"] + args, check=True)
+
+
+def git_commit(message):
+    """Git commit."""
+    run(["git", "commit", "--allow-empty", "-m", message], check=True)
 
 
 def git_clone(source, destination):
@@ -163,7 +265,7 @@ def checkout(rev):
 
 
 def is_ancestor(rev, last):
-    """Check if revisiion can fast-forward"""
+    """Check if revisiion can fast-forward."""
     try:
         run(["git", "merge-base", "--is-ancestor", last, rev], check=True)
         return True
@@ -425,6 +527,26 @@ def less_boring():
     disable.rename(bfile)
 
 
+@contextmanager
+def ignore_darcs():
+    """Replace .gitignore with one that ignores `_darcs`."""
+    bfile = Path(".gitignore")
+    old = None
+    if bfile.exists():
+        with bfile.open("rb") as f:
+            old = f.read()
+        bfile.unlink()
+
+    with bfile.open("w", encoding="UTF-8") as f:
+        f.write(_gitignore)
+    yield
+    bfile.unlink()
+    if old:
+        with bfile.open("wb") as f:
+            f.write(old)
+        git_add([".gitignore"])
+
+
 def transfer(gen, count, *, last=None):
     """Transfer the git-commits to darcs."""
     try:
@@ -502,61 +624,8 @@ def setup(warn, verbose):
         _disable = True
 
 
-@click.group()
-def main():
-    """Click entrypoint."""
-    fix_pwd()
-
-
-@main.command()
-@click.argument("source", type=click.Path(exists=True, dir_okay=True, file_okay=False))
-@click.argument("destination", type=click.Path(exists=False))
-@click.option("-v/-nv", "--verbose/--no-verbose", default=False)
-def clone(source, destination, verbose):
-    """Locally clone a tracking repository to get a working-repository."""
-    setup(False, verbose=verbose)
-    destination = Path(destination)
-    if destination.exists():
-        raise ClickException(f"Destination `{destination}` may not exist")
-    git_clone(source, destination)
-    darcs_dest = Path(destination, _uuid)
-    repo_source = Path(darcs_dest, "_darcs")
-    darcs_clone(source, darcs_dest)
-    repo_source = Path(darcs_dest, "_darcs")
-    repo_dest = Path(destination, "_darcs")
-    repo_source.rename(repo_dest)
-    rmtree(darcs_dest, ignore_errors=True)
-    os.chdir(destination)
-    relink()
-
-
-@main.command()
-@click.option("-v/-nv", "--verbose/--no-verbose", default=False)
-@click.option(
-    "-w/-nw",
-    "--warn/--no-warn",
-    default=True,
-    help="Warn that repository will be cleared",
-)
-@click.option(
-    "--base",
-    "-b",
-    default=None,
-    help="On first update import from (commit-ish)",
-)
-@click.option(
-    "-s/-ns",
-    "--shallow/--no-shallow",
-    default=None,
-    help="On first update only import current commit",
-)
-def update(verbose, base, warn, shallow):
-    """Incremental import of git into darcs.
-
-    By default it imports a shallow copy (the current commit). Use `--no-shallow`
-    to import the complete history.
-    """
-    setup(warn, verbose=verbose)
+def prepare_update(base, shallow):
+    """Check repo, arguments, find base and setup shutdown."""
     if not Path(".git").exists():
         raise ClickException("Please run git-darcs in the root of your git-repo.")
     if not Path("_darcs").exists():
@@ -583,6 +652,11 @@ def update(verbose, base, warn, shallow):
     if _isatty:
         _thread = Thread(target=handle_shutdown, daemon=True)
         _thread.start()
+    return rbase, from_checkpoint, do_one
+
+
+def run_update(rbase, from_checkpoint, do_one):
+    """Run conversion loop."""
     branch = get_current_branch()
     failed = True
     try:
@@ -602,3 +676,287 @@ def update(verbose, base, warn, shallow):
             else:
                 wipe()
                 checkout(branch)
+
+
+def ask(question, choice, *, text="", state="", help=""):
+    """Ask a question on the terminal."""
+    key = "?"
+    if text:
+        print(text)
+    while key in ("h", "?"):
+        sys.stdout.write(f"{question}? {state}  [{choice}], or ? for more options: ")
+        sys.stdout.flush()
+        key = readkey()
+        print(key)
+        if key in ("h", "?"):
+            print(help)
+    return key
+
+
+class Patch:
+    """Represents a darcs-patch."""
+
+    def __init__(self, source, patch):
+        """Dear flake8 this is a init function."""
+        self.source = source
+        self.patch = patch
+        fields = patch.attrib
+        self.author = fields["author"]
+        self.hash = fields["hash"]
+        self.chash = Fore.CYAN + f"patch {self.hash}" + Style.RESET_ALL
+        self.date = datetime.strptime(fields["date"], _darcs_date)
+        self.subject = patch.find("name").text
+        comment = patch.find("comment").text
+        if comment.startswith("Ignore-this: "):
+            comment = os.linesep.join(comment.splitlines()[1:])
+        self.comment = comment.strip()
+        self.pull = None
+
+    def short(self):
+        """Get a short description of the patch."""
+        return f"""
+{self.chash}
+Author: {self.author}
+Date: {self.date}
+Subject: {self.subject}
+        """.strip()
+
+    def long(self):
+        """Get a longer description of the patch."""
+        if self.comment:
+            return f"{self.short()}\n\n{self.comment}"
+        else:
+            return self.short()
+
+    def full(self):
+        """Show the full patch."""
+        show_full_patch(self.source, self.hash)
+
+    def message(self):
+        """Format patch message for git."""
+        return f"{self.subject}\n\n{self.comment}".strip()
+
+    def ask(self, index, of):
+        """Ask if I should pull this patch."""
+        key = "l"
+        text = self.short()
+        while key in ("l", "f"):
+            key = ask(
+                _pull_question,
+                "ynwasc",
+                text=text,
+                state=f"{index + 1}/{of}",
+                help=_pull_help,
+            )
+            text = ""
+            if key == "l":
+                print(self.long())
+            elif key == "f":
+                self.full()
+            elif key == "y":
+                self.pull = True
+            elif key == "n":
+                self.pull = False
+            elif key == "a":
+                self.pull = True
+            elif key == "i":
+                self.pull = False
+        return key
+
+
+class Pull:
+    """Contains state for the pull-questions."""
+
+    def __init__(self, source, args, *, ignore_temp=True):
+        """Dear flake8 this is a init function."""
+        self.source = source
+        self.args = args
+        self.ignore_temp = ignore_temp
+        self.patches_xml = get_patches(source, args)
+        self.patches = OrderedDict()  # Legacy support
+        for patch in self.patches_xml:
+            obj = Patch(source, patch)
+            if self.ignore_temp:
+                if not obj.subject.startswith("temp: "):
+                    self.patches[obj.hash] = obj
+            else:
+                self.patches[obj.hash] = obj
+
+    def pull(self, all=False):
+        """Pull patches."""
+        if not self.patches:
+            print("No remote patches to pull in!")
+            return
+        if all:
+            for patch in self.patches.values():
+                patch.pull = True
+        else:
+            self.decide()
+        pull = [x for x in self.patches.values() if x.pull]
+        count = len(pull)
+        if not all:
+            key = ask(f"Shall I pull {count} patches", "yn")
+            if key in ("n", "q", "c"):
+                print("Cancel pull")
+                sys.exit(1)
+        with tqdm(desc="pull", total=count, disable=_disable) as pbar:
+            for patch in pull:
+                pull_patch(self.source, patch.hash)
+                with ignore_darcs():
+                    git_add()
+                git_commit(patch.message())
+                pbar.update()
+
+    def pull_depends(self, hash):
+        """Find dependent patches an set pull to True for these, too."""
+        xml = get_patches(self.source, ["-h", hash])
+        count = 0
+        for patch in xml:
+            hash = patch.attrib["hash"]
+            patch = self.patches[hash]
+            if not patch.pull:
+                count += 1
+                patch.pull = True
+        if count:
+            print(Fore.RED + f"Select {count} dependent patches" + Style.RESET_ALL)
+
+    def decide(self):
+        """Decide patches to pull."""
+        decide = OrderedDict(self.patches)
+        while decide:
+            key = None
+            of = len(decide)
+            for index, (hash, patch) in enumerate(OrderedDict(decide).items()):
+                key = patch.ask(index, of)
+                if patch.pull is not None:
+                    if patch.pull:
+                        self.pull_depends(hash)
+                    decide.pop(hash)
+                if key == "a":
+                    break
+                elif key == "i":
+                    break
+                elif key in ("c", "q"):
+                    print("Cancel pull")
+                    sys.exit(1)
+
+            of = len(decide)
+            with tqdm(desc="resolve", total=of, disable=_disable) as pbar:
+                for hash, patch in OrderedDict(decide).items():
+                    if key == "a":
+                        patch.pull = True
+                        self.pull_depends(hash)
+                    elif key == "i":
+                        patch.pull = False
+                    decide.pop(hash)
+                    pbar.update()
+
+
+@click.group()
+def main():
+    """Click entrypoint."""
+    fix_pwd()
+
+
+@main.command()
+@click.argument("source", type=click.Path(exists=True, dir_okay=True, file_okay=False))
+@click.argument("destination", type=click.Path(exists=False))
+@click.option("-v/-nv", "--verbose/--no-verbose", default=False)
+def clone(source, destination, verbose):
+    """Locally clone a tracking-repository to get a working-repository."""
+    setup(False, verbose=verbose)
+    with tqdm(desc="clone", total=5, disable=_disable) as pbar:
+        destination = Path(destination)
+        if destination.exists():
+            raise ClickException(f"Destination `{destination}` may not exist")
+        git_clone(source, destination)
+        pbar.update()
+        dest = Path(destination, ".git", "config")
+        dest.unlink()
+        copy(Path(source, ".git", "config"), dest)
+        pbar.update()
+        darcs_dest = Path(destination, _uuid)
+        repo_source = Path(darcs_dest, "_darcs")
+        darcs_clone(source, darcs_dest)
+        pbar.update()
+        repo_source = Path(darcs_dest, "_darcs")
+        repo_dest = Path(destination, "_darcs")
+        repo_source.rename(repo_dest)
+        rmtree(darcs_dest, ignore_errors=True)
+        pbar.update()
+        os.chdir(destination)
+        relink()
+        pbar.update()
+
+
+@main.command()
+@click.option("-v/-nv", "--verbose/--no-verbose", default=False)
+@click.option(
+    "-w/-nw",
+    "--warn/--no-warn",
+    default=True,
+    help="Warn that repository will be cleared",
+)
+@click.option(
+    "--base",
+    "-b",
+    default=None,
+    help="On first update import from (commit-ish)",
+)
+@click.option(
+    "-s/-ns",
+    "--shallow/--no-shallow",
+    default=None,
+    help="On first update only import current commit",
+)
+def update(verbose, warn, base, shallow):
+    """Incremental import of git into darcs.
+
+    By default it imports a shallow copy (the current commit). Use `--no-shallow`
+    to import the complete history.
+    """
+    setup(warn, verbose=verbose)
+    run_update(*prepare_update(base, shallow))
+
+
+@main.command()
+@click.option("-v/-nv", "--verbose/--no-verbose", default=False)
+@click.option(
+    "-w/-nw",
+    "--warn/--no-warn",
+    default=True,
+    help="Warn that repository will be cleared",
+)
+@click.option(
+    "-a/-na",
+    "--all/--no-all",
+    default=False,
+    help="Pull all patches",
+)
+@click.option(
+    "-i/-ni",
+    "--ignore-temp/--no-ignore-temp",
+    default=True,
+    help="Ignore temporary patches (with 'temp: ')",
+)
+@click.argument("source", type=click.Path(exists=True, dir_okay=True, file_okay=False))
+@click.argument("darcs", nargs=-1)
+def pull(verbose, all, warn, source, darcs, ignore_temp):
+    """Pull from source darcs-repository into a tracking-repository.
+
+    A tracking-repository is created by `git darcs update` and contains a git- and a
+    darcs-repository. Arguments after `--` are passed to `darcs pull`.
+    """
+    setup(warn, verbose=verbose)
+    if not Path("_darcs").exists():
+        raise ClickException("Please run git-darcs in the root of your darcs-repo.")
+    if not Path(".git").exists():
+        raise ClickException("Please run git-darcs in the root of your git-repo.")
+    wipe()
+    if hasnew():
+        raise ClickException(
+            "The git and the darcs repo in your tracking-repo are not in sync."
+        )
+
+    init()
+    Pull(source, list(darcs), ignore_temp=ignore_temp).pull(all)
